#include <pgmspace.h>

#define THINGNAME "MyESP32CAM_0"

const char WIFI_SSID[] = "QQ";
const char WIFI_PASSWORD[] = "12345678";
const char SCHOOL_IOT_ENDPOINT[] = "iot.cpe.ku.ac.th";
const int Led_Flash = 4;

// Amazon Root CA 1
static const char AWS_CERT_CA[] PROGMEM = R"EOF(
-----BEGIN CERTIFICATE-----
-----END CERTIFICATE-----
)EOF";

// Device Certificate
static const char AWS_CERT_CRT[] PROGMEM = R"KEY(
-----BEGIN CERTIFICATE-----
-----END CERTIFICATE-----
)KEY";

// Device Private Key
static const char AWS_CERT_PRIVATE[] PROGMEM = R"KEY(
-----BEGIN RSA PRIVATE KEY-----
-----END RSA PRIVATE KEY-----
)KEY";


#include <WiFiClientSecure.h>
#include <MQTTClient.h>
#include "WiFi.h"
#include "esp_camera.h"
#include "esp_system.h"
#include <PubSubClient.h>
#include <TridentTD_LineNotify.h>

#define LINE_TOKEN  "gCpuLec5ucVH7ZHXPWk3BM0dJrpkuvFWZNzMXhlfCqg" // TOKEN

#define PWDN_GPIO_NUM 32
#define RESET_GPIO_NUM -1
#define XCLK_GPIO_NUM 0
#define SIOD_GPIO_NUM 26
#define SIOC_GPIO_NUM 27
#define Y9_GPIO_NUM 35
#define Y8_GPIO_NUM 34
#define Y7_GPIO_NUM 39
#define Y6_GPIO_NUM 36
#define Y5_GPIO_NUM 21
#define Y4_GPIO_NUM 19
#define Y3_GPIO_NUM 18
#define Y2_GPIO_NUM 5
#define VSYNC_GPIO_NUM 25
#define HREF_GPIO_NUM 23
#define PCLK_GPIO_NUM 22
#define TOPIC_SW   "b6510503247/switch"
#define ESP32CAM_PUBLISH_TOPIC   "b6510503247/esp32/cam_0"
WiFiClient wifiClient;
PubSubClient mqtt(SCHOOL_IOT_ENDPOINT, 1883, wifiClient);

const int bufferSize = 1024 * 23;  // 23552 bytes


void connectAWS() {
  WiFi.mode(WIFI_STA);
  WiFi.begin(WIFI_SSID, WIFI_PASSWORD);

  Serial.println("\n\n=====================");
  Serial.println("Connecting to Wi-Fi");
  Serial.println("=====================\n\n");

  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }

  // Configure WiFiClientSecure to use the AWS IoT device credentials
  //net.setCACert(AWS_CERT_CA);
  // net.setCertificate(AWS_CERT_CRT);
  // net.setPrivateKey(AWS_CERT_PRIVATE);
  // Connect to the MQTT broker on the AWS endpoint we defined earlier

  Serial.println("\n\n=====================");
  Serial.println("Connecting to SCHOOL IOT");
  Serial.println("=====================\n\n");
  
  connect_mqtt();

  while (!mqtt.connect(THINGNAME)) {
    Serial.print(".");
    delay(100);
  }

  if (!mqtt.connected()) {
    Serial.println("SCHOOL IoT Timeout!");
    ESP.restart();
    return;
  }

  Serial.println("\n\n=====================");
  Serial.println("SCHOOL IoT Connected!");
  Serial.println("=====================\n\n");
}

void cameraInit() {
  camera_config_t config;
  config.ledc_channel = LEDC_CHANNEL_0;
  config.ledc_timer = LEDC_TIMER_0;
  config.pin_d0 = Y2_GPIO_NUM;
  config.pin_d1 = Y3_GPIO_NUM;
  config.pin_d2 = Y4_GPIO_NUM;
  config.pin_d3 = Y5_GPIO_NUM;
  config.pin_d4 = Y6_GPIO_NUM;
  config.pin_d5 = Y7_GPIO_NUM;
  config.pin_d6 = Y8_GPIO_NUM;
  config.pin_d7 = Y9_GPIO_NUM;
  config.pin_xclk = XCLK_GPIO_NUM;
  config.pin_pclk = PCLK_GPIO_NUM;
  config.pin_vsync = VSYNC_GPIO_NUM;
  config.pin_href = HREF_GPIO_NUM;
  config.pin_sscb_sda = SIOD_GPIO_NUM;
  config.pin_sscb_scl = SIOC_GPIO_NUM;
  config.pin_pwdn = PWDN_GPIO_NUM;
  config.pin_reset = RESET_GPIO_NUM;
  config.xclk_freq_hz = 20000000;
  config.pixel_format = PIXFORMAT_JPEG;
  config.frame_size = FRAMESIZE_VGA;  // 640x480
  config.jpeg_quality = 10;
  config.fb_count = 2;

  // camera init
  esp_err_t err = esp_camera_init(&config);
  if (err != ESP_OK) {
    Serial.printf("Camera init failed with error 0x%x", err);
    ESP.restart();
    return;
  }
}

void grabImage() {
  camera_fb_t* fb = esp_camera_fb_get();
  if (fb != NULL && fb->format == PIXFORMAT_JPEG && fb->len < bufferSize) {
    Serial.print("\nImage Length: ");
    Serial.print(fb->len);
    Serial.print("\t Image: ");
    delay(1000);
    Serial.print((const char*)fb->buf);
    // bool result = mqtt.publish(ESP32CAM_PUBLISH_TOPIC, "hello world");
    bool result = mqtt.publish(ESP32CAM_PUBLISH_TOPIC, (const char*)fb->buf, fb->len);
    
    Serial.print("\n result ");
    Serial.print(result);

    if (!result) {
      ESP.restart();
    }
  }
  esp_camera_fb_return(fb);
  delay(1);
}

void setup() {
  Serial.begin(115200);
  pinMode(Led_Flash, OUTPUT);
  cameraInit();
  connectAWS();
  LINE.setToken(LINE_TOKEN);
}

void loop() {
  mqtt.loop();
}

bool debounce = false;
void mqtt_callback(char* topic, byte* payload, unsigned int length) {
  if (strcmp(topic, TOPIC_SW) == 0) {
    payload[length] = 0; // null-terminate the payload to treat it as a string
    int value = atoi((char*)payload); 
    if (value == 0 && debounce) {
      debounce = false;
    }
    else if (value == 1 && !debounce) {
      Camera_capture();
      printf("HERE.\n");
      debounce = true;
    }
    else {
      printf("Invalid payload received.\n");
    }
  }
}
void connect_mqtt() {
  printf("Connecting to MQTT broker at %s.\n", SCHOOL_IOT_ENDPOINT);
  if (!mqtt.connect("", "b6510503247", "kanathip.p@ku.th")) {
    printf("Failed to connect to MQTT broker.\n");
    for (;;) {} // wait here forever
  }
  mqtt.setCallback(mqtt_callback);
  mqtt.subscribe(TOPIC_SW);
  printf("MQTT broker connected.\n");
}

void Send_line(uint8_t *image_data, size_t   image_size) {
  LINE.notifyPicture("ภาพที่ได้จากกล้อง", image_data, image_size);
}


void Camera_capture() {
  digitalWrite(Led_Flash, HIGH);
  delay(100);
  digitalWrite(Led_Flash, LOW);
  delay(100);
  digitalWrite(Led_Flash, HIGH);
  camera_fb_t * fb = NULL;
  delay(200);
  // Take Picture with Camera
  fb = esp_camera_fb_get();
  if (!fb) {
    Serial.println("Camera capture failed");
    return;
  }
  digitalWrite(Led_Flash, LOW);
  Send_line(fb->buf, fb->len);
  esp_camera_fb_return(fb);
  // Serial.println("Going to sleep now");
  // esp_deep_sleep_start();
  // Serial.println("This will never be printed");

}